from pycomposefile.compose_element import ComposeElement, ComposeListOrMapElement


class Options(ComposeListOrMapElement):
    pass


class Logging(ComposeElement):
    element_keys = {
        "driver": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#logging"),
        "options": (Options, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#logging")
    }



import unittest

from ..compose_generator import ComposeGenerator


class TestComposeServiceSecretsFile(unittest.TestCase):
    def test_service_with_single_secret(self):
        secret_value = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.\n"

        compose_file = ComposeGenerator.get_compose_with_one_secret()
        secret_from_file = compose_file.secrets["my_secret"].file.readFile()

        self.assertEqual(compose_file.services["frontend"].secrets[0].source, "my_secret")
        self.assertEqual(compose_file.secrets["my_secret"].file, "./sample/my_secret.txt")
        self.assertEqual(secret_from_file, secret_value)

    def test_service_with_external_secret(self):
        compose_file = ComposeGenerator.get_compose_with_external_secret()

        self.assertEqual(compose_file.services["frontend"].secrets[0].source, "my_other_secret")
        self.assertEqual(compose_file.secrets["my_other_secret"].external, True)

    def test_service_with_long_syntax(self):
        secret_value = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.\n"

        compose_file = ComposeGenerator.get_compose_with_long_syntax()
        secret_from_file = compose_file.secrets["my_secret"].file.readFile()

        self.assertEqual(compose_file.services["frontend"].secrets[0].source, "my_secret")
        self.assertEqual(compose_file.services["frontend"].secrets[0].target, "redis_secret")
        self.assertEqual(compose_file.services["frontend"].secrets[0].uid, 103)
        self.assertEqual(compose_file.services["frontend"].secrets[0].gid, 103)
        self.assertEqual(compose_file.secrets["my_secret"].file, "./sample/my_secret.txt")
        self.assertEqual(secret_from_file, secret_value)



from unittest import TestCase, mock
import os
from ..compose_generator import ComposeGenerator
from pycomposefile.compose_element import EmptyOrUnsetException


class TestBracesNoUnderscoreNoDigitVariableInterpolation(TestCase):

    @mock.patch.dict(os.environ, {"TESTNAME": "fred"})
    def test_uppercase_in_string_value(self):
        braced_env_var = "{TESTNAME}"
        compose_file = ComposeGenerator.get_compose_with_string_value(braced_env_var)
        self.assertEqual(compose_file.services["frontend"].image, "awesome/fred")

    def test_uppercase_with_default_when_unset_in_string_value(self):
        env_var = "DEFAULTUNSET"
        os.unsetenv(env_var)
        braced_env_with_default_unset = "{" + env_var + ":-bob}"
        compose_file = ComposeGenerator.get_compose_with_string_value(braced_env_with_default_unset)
        self.assertEqual(compose_file.services["frontend"].image, "awesome/bob")

    @mock.patch.dict(os.environ, {"TESTCPUCOUNT": "1.5"})
    def test_uppercase_in_decimal_value(self):
        braced_env_var = "{TESTCPUCOUNT}"
        compose_file = ComposeGenerator.get_compose_with_decimal_value(braced_env_var)
        self.assertEqual(compose_file.services["frontend"].cpu_count, 1.5)

    @mock.patch.dict(os.environ, {"HOSTPORT": "8080"})
    @mock.patch.dict(os.environ, {"CONTAINERPORT": "80"})
    def test_uppercase_two_variables_in_string_value(self):
        braced_first_env_var = "{HOSTPORT}"
        braced_second_env_var = "{CONTAINERPORT}"
        compose_file = ComposeGenerator.get_with_two_environment_variables_in_string_value(braced_first_env_var, braced_second_env_var)
        self.assertEqual(f"{compose_file.services['frontend'].ports[0]}", "8080:80/tcp")
        self.assertEqual(compose_file.services['frontend'].ports[0].published, "8080")
        self.assertEqual(compose_file.services['frontend'].ports[0].target, "80")

    @mock.patch.dict(os.environ, {"testname": "fred"})
    def test_lowercase_in_string_value(self):
        braced_env_var = "{testname}"
        compose_file = ComposeGenerator.get_compose_with_string_value(braced_env_var)
        self.assertEqual(compose_file.services["frontend"].image, "awesome/fred")

    def test_lowercase_with_default_when_unset_in_string_value(self):
        env_var = "defaultunset"
        os.unsetenv(env_var)
        braced_env_with_default_unset = "{" + env_var + ":-bob}"
        compose_file = ComposeGenerator.get_compose_with_string_value(braced_env_with_default_unset)
        self.assertEqual(compose_file.services["frontend"].image, "awesome/bob")

    @mock.patch.dict(os.environ, {"testcpucount": "1.5"})
    def test_lowercase_in_decimal_value(self):
        braced_env_var = "{testcpucount}"
        compose_file = ComposeGenerator.get_compose_with_decimal_value(braced_env_var)
        self.assertEqual(compose_file.services["frontend"].cpu_count, 1.5)

    @mock.patch.dict(os.environ, {"hostport": "8080"})
    @mock.patch.dict(os.environ, {"containerport": "80"})
    def test_lowercase_two_variables_in_string_value(self):
        braced_first_env_var = "{hostport}"
        braced_second_env_var = "{containerport}"
        compose_file = ComposeGenerator.get_with_two_environment_variables_in_string_value(braced_first_env_var, braced_second_env_var)
        self.assertEqual(f"{compose_file.services['frontend'].ports[0]}", "8080:80/tcp")
        self.assertEqual(compose_file.services["frontend"].ports[0].published, "8080")
        self.assertEqual(compose_file.services["frontend"].ports[0].target, "80")

    @mock.patch.dict(os.environ, {"RACK_ENV": "test"})
    @mock.patch.dict(os.environ, {"test": "https://127.0.0.1"})
    @mock.patch.dict(os.environ, {"VERSION": "release"})
    def test_service_with_os_environment_vars(self):
        compose_file = ComposeGenerator.get_compose_with_os_environment_vars()

        self.assertEqual(compose_file.services["frontend"].environment["RACK_ENV"], "test")
        self.assertEqual(compose_file.services["frontend"].environment["URL"], "https://127.0.0.1")
        self.assertEqual(compose_file.services["frontend"].environment["VERSION"], "release")

    @mock.patch.dict(os.environ, {"URL": "https://127.0.0.1"})
    def test_service_with_mandatory_env_vars(self):
        with self.assertRaises(EmptyOrUnsetException):
            ComposeGenerator.get_compose_with_mandatory_env_vars()

    def test_service_with_mandatory_unset_env_vars(self):
        with self.assertRaises(EmptyOrUnsetException):
            ComposeGenerator.get_compose_with_mandatory_unset_env_vars()

    @mock.patch.dict(os.environ, {"ENVIRONMENT": "local"})
    def test_service_with_double_dollar_sign_env_vars(self):
        compose_file = ComposeGenerator.get_compose_with_double_dollar_sign_env_vars()

        self.assertEqual(compose_file.services["frontend"].environment["ENVIRONMENT"], "$ENVIRONMENT")



from pycomposefile.compose_element import ComposeElement


class SecretFile(str):
    def readFile(self):
        with open(self, "r") as f:
            secret = f.read()
        return secret


class Secrets(ComposeElement):
    element_keys = {
        "name": (str, ""),
        "file": (SecretFile, ""),
        "external": (bool, ""),
    }



from pycomposefile.compose_element import ComposeListOrMapElement, ComposeStringOrListElement, ComposeElement


class Network(ComposeElement):
    element_keys = {
        "aliases": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#aliases"),
        "ipv4_address": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#ipv4_address-ipv6_address"),
        "ipv6_address": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#ipv4_address-ipv6_address"),
        "link_local_ips": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#link_local_ips"),
        "priority": (int, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#priority"),
    }


class Networks(ComposeListOrMapElement):
    transform = Network.from_parsed_yaml



import re
import os


class EmptyOrUnsetException(Exception):
    def __init__(self, variable_name, *args: object) -> None:
        self.variable_name = variable_name
        super().__init__(*args)

    def __str__(self) -> str:
        return f"Failed to evaluate mandatory variable {self.variable_name}"


class ComposeDataTypeTransformer():
    transform = str
    valid_values = None

    def transform_supported_data(self, value):
        if self.valid_values is not None:
            value = self.transform_and_validate_supported_data(value)
        else:
            value = self.transform(self.replace_environment_variables(value))
        return value

    def test_valid_string(self, value):
        if isinstance(value, str):
            return value in self.valid_values
        return False

    def test_valid_int(self, value):
        if isinstance(value, int):
            return self.valid_values[0] <= value <= self.valid_values[1]
        return False

    def transform_and_validate_supported_data(self, value):
        # TODO: if the data is an integer or decimal, should there be a "between" check?
        transformed = self.transform(self.replace_environment_variables(value))
        if self.test_valid_string(transformed) or self.test_valid_int(transformed):
            return transformed
        else:
            # TODO: Should this return None or should this raise?
            return None

    def replace_environment_variables(self, value):
        value = str(value)
        value = self.replace_environment_variables_with_braces(value)
        value = self.replace_environment_variables_without_braces(value)
        value = self.replace_environment_variables_with_empty_unset(value)
        value = self.replace_environment_variables_with_unset(value)
        value = self.replace_mandatory_variables_with_empty_or_unset(value)
        value = self.replace_mandatory_variables_with_unset(value)
        value = self.do_not_replace_variables(value)

        return value

    def update_value_with_resolved_environment(self, env_variable_name, env_variable_value, source_string):
        if env_variable_value is None:
            env_variable_value = ""
        return re.sub(env_variable_name, env_variable_value, source_string)

    def replace_environment_variables_with_empty_unset(self, value):
        capture = re.compile(r"(\$+)\{(?P<variablename>\w+)\:-(?P<defaultvalue>.+)\}")
        matches = capture.search(value)
        if matches is not None and matches[1] == "$$":
            return value
        while matches:
            env_var = os.environ.get(matches.group("variablename"))
            default_value = matches.group("defaultvalue")
            if env_var is None or len(env_var) == 0:
                env_var = default_value
            value = self.update_value_with_resolved_environment(f"\\{matches[0]}", env_var, value)
            matches = capture.search(value)
        return value

    def replace_environment_variables_with_unset(self, value):
        capture = re.compile(r"(\$+)\{(?P<variablename>\w+)-(?P<defaultvalue>.+)\}")
        matches = capture.search(value)
        if matches is not None and matches[1] == "$$":
            return value
        while matches:
            env_var = os.environ.get(matches.group("variablename"))
            default_value = matches.group("defaultvalue")
            if env_var is None:
                env_var = default_value
            value = self.update_value_with_resolved_environment(f"{matches[0]}", env_var, value)
            matches = capture.search(value)
        return value

    def replace_mandatory_variables_with_empty_or_unset(self, value):
        capture = re.compile(r"(\$+)\{(?P<variablename>\w+)\:\?(err)\}")
        matches = capture.search(value)
        if matches is not None and matches[1] == "$$":
            return value
        while matches:
            env_var = os.environ.get(matches.group("variablename"))
            if env_var is None or len(env_var) == 0:
                raise EmptyOrUnsetException(matches.group("variablename"))
            to_be_replaced = r"\$\{" + matches.group('variablename') + r"\:\?err\}"
            value = self.update_value_with_resolved_environment(to_be_replaced, env_var, value)
            matches = capture.search(value)
        return value

    def replace_mandatory_variables_with_unset(self, value):
        capture = re.compile(r"(\$+)\{(?P<variablename>\w+)\?(err)\}")
        matches = capture.search(value)
        if matches is not None and matches[1] == "$$":
            return value
        while matches:
            env_var = os.environ.get(matches.group("variablename"))
            if env_var is None:
                raise EmptyOrUnsetException(matches.group("variablename"))
            to_be_replaced = r"\$\{" + matches.group('variablename') + r"\?err\}"
            value = self.update_value_with_resolved_environment(to_be_replaced, env_var, value)
            matches = capture.search(value)
        return value

    def replace_environment_variables_with_braces(self, value):
        capture = re.compile(r"\$\{(?P<variablename>\w+)\}")
        matches = capture.search(value)
        if matches is not None and matches[1] == "$$":
            return value
        while matches:
            env_var = os.environ.get(matches.group("variablename"))
            value = self.update_value_with_resolved_environment(f"\\{matches[0]}", env_var, value)
            matches = capture.search(value)
        return value

    def do_not_replace_variables(self, value):
        value = re.sub(r"\$\$", "$", value)
        return value

    def replace_environment_variables_without_braces(self, value):
        capture = re.compile(r"(\$+)(?P<variablename>\w+)")
        matches = capture.search(value)
        if matches is not None and matches[1] == "$$":
            return value
        while matches:
            env_var = os.environ.get(matches.group("variablename"))
            value = self.update_value_with_resolved_environment(f"\\{matches[0]}", env_var, value)
            matches = capture.search(value)
        return value



from pycomposefile.compose_element import (ComposeElement,
                                           ComposeStringOrListElement,
                                           ComposeByteValue)


class Tmpfs(ComposeElement):
    element_keys = {
        "size": (ComposeByteValue, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "mode": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
    }


class Bind(ComposeElement):
    element_keys = {
        "propigation": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "create_host_path": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "selinux": ((str, ['z', 'Z']), "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
    }


class Volume(ComposeElement):
    element_keys = {
        "nocopy": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4")
    }


class VolumeMap(ComposeElement):
    element_keys = {
        "type": ((str, ["volume", "bind", "tmpfs", "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"]), "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "source": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "target": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "volume": (Volume.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "bind": (Bind.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "read_only": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "tmpfs": (Tmpfs.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
        "consistency": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-4"),
    }

    @classmethod
    def from_parsed_yaml(cls, config, name, compose_path):
        if config is None:
            return None
        if isinstance(config, str):
            new_dict = {}
            split_string = config.split(":")
            if len(split_string) == 3:
                new_dict["source"], new_dict["target"], access_mode = split_string
                access_mode = access_mode.split(',')

                if 'rw' in access_mode:
                    new_dict["read_only"] = False
                elif "r" in access_mode:
                    new_dict["read_only"] = True

                if 'z' in access_mode:
                    bind = {}
                    bind["selinux"] = 'z'
                    new_dict["bind"] = bind
                elif 'Z' in access_mode:
                    bind = {}
                    bind["selinux"] = 'Z'
                    new_dict["bind"] = bind
            else:
                new_dict["source"], new_dict["target"], options = split_string
            config = new_dict
        compose_path = f"{compose_path}/{name}"
        return cls(config, compose_path)


class Volumes(ComposeStringOrListElement):
    transform = VolumeMap.from_parsed_yaml



from unittest import TestCase

from src.pycomposefile import compose_file

from ..compose_generator import ComposeGenerator


class TestComposeServiceEnvironment(TestCase):
    def test_service_with_environment_map(self):
        compose_file = ComposeGenerator.get_compose_with_environment_map()

        self.assertEqual(compose_file.services["frontend"].environment["RACK_ENV"], "development")
        self.assertEqual(compose_file.services["frontend"].environment["SHOW"], "true")
        self.assertIsNone(compose_file.services["frontend"].environment["USER_INPUT"])

    def test_service_with_environment_list(self):
        compose_file = ComposeGenerator.get_compose_with_environment_list()

        self.assertEqual(compose_file.services["frontend"].environment["RACK_ENV"], "development")
        self.assertEqual(compose_file.services["frontend"].environment["SHOW"], "true")
        self.assertIsNone(compose_file.services["frontend"].environment["USER_INPUT"])



from .compose_datatype_transformer import ComposeDataTypeTransformer


class ComposeListOrMapElement(ComposeDataTypeTransformer, dict):
    def isValueEmpty(self, line):
        line = line.rstrip("=")
        if "=" not in line:
            return True
        else:
            return False

    def evaluate_from_string(self, source_string):
        source_string = self.replace_environment_variables(source_string)
        if self.isValueEmpty(source_string):
            value = None
            key = source_string.rstrip("=")
        else:
            key, value = source_string.split("=")
        self[key] = value

    def evaluate_from_list(self, source_list):
        for line in source_list:
            self.evaluate_from_string(line)

    def evaluate_from_dict(self, config, compose_path):
        for key in config.keys():
            value = config[key]
            if value is None:
                pass
            elif isinstance(value, dict):
                value = self.transform(value, key, compose_path)
            elif isinstance(value, list):
                new_list = []
                for v in value:
                    new_list.append(self.replace_environment_variables(v))
                value = new_list
            else:
                value = self.replace_environment_variables(value)
            self[key] = value

    def __init__(self, config, key=None, compose_path=None,):
        if compose_path is not None:
            self.compose_path = f"{compose_path}/{key}"
        if config is None:
            pass
        else:
            if isinstance(config, dict):
                self.evaluate_from_dict(config, compose_path)
            elif isinstance(config, list):
                self.evaluate_from_list(config)
            elif isinstance(config, str):
                self.evaluate_from_string(config)



from decimal import Decimal
from tkinter import Label
from unittest.loader import VALID_MODULE_NAME

from pycomposefile.service.service_misc import (Expose,
                                                DependsOn,
                                                StorageOpt,
                                                Ulimits)
from pycomposefile.service.service_blkio_config import BlkioConfig
from pycomposefile.service.service_build import Build
from pycomposefile.service.service_deploy import Deploy
from pycomposefile.service.service_credential_spec import CredentialSpec
from pycomposefile.service.service_cap import Cap
from pycomposefile.service.service_configs import (Configs, Secrets)
from pycomposefile.service.service_command import Command
from pycomposefile.service.service_environment import (Environment, EnvFile)
from pycomposefile.service.service_healthcheck import HealthCheck
from pycomposefile.service.service_logging import Logging
from pycomposefile.service.service_networks import Networks
from pycomposefile.service.service_ports import Ports
from pycomposefile.service.service_volumes import Volumes
from pycomposefile.compose_element import (ComposeElement,
                                           ComposeListOrMapElement,
                                           ComposeByteValue,
                                           ComposeStringOrListElement)


class Service(ComposeElement):
    element_keys = {
        "image": (str, ""),
        "container_name": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#container_name"),
        "cpu_count": (Decimal, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_count"),
        "entrypoint": (Command, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#entrypoint"),
        "command": (Command, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#command"),
        "deploy": (Deploy.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/deploy.md"),
        "expose": (Expose, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#expose"),
        "ports": (Ports, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#long-syntax-2"),
        "cpus": (Decimal, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpus"),
        "credential_spec": (CredentialSpec.from_parsed_yaml, ""),
        "blkio_config": (BlkioConfig.from_parsed_yaml,
                         "https://github.com/compose-spec/compose-spec/blob/master/spec.md#blkio_config"),
        "cpu_percent": (Decimal,
                        "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_percent"),
        "cpu_shares": (int,
                       "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_shares"),
        "cpu_period": (str,

                       "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_period"),
        "cpu_quota": (int,
                      "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_quota"),
        "cpu_rt_runtime": (str,
                           "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_rt_runtime"),
        "cpu_rt_period": (str,
                          "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpu_rt_period"),
        "cpuset": (ComposeStringOrListElement,
                   "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cpuset"),
        "build": (Build.from_parsed_yaml,
                  "https://github.com/compose-spec/compose-spec/blob/master/build.md"),
        "cap_add": (Cap,
                    "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cap_add"),
        "cap_drop": (Cap,
                     "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cap_add"),
        "cgroup_parent": (str,
                          "https://github.com/compose-spec/compose-spec/blob/master/spec.md#cgroup_parent"),
        "configs": (Configs,
                    "https://github.com/compose-spec/compose-spec/blob/master/spec.md#configs"),
        "depends_on": (DependsOn, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#depends_on"),
        "env_file": (EnvFile, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#env_file"),
        "environment": (Environment, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#environment"),
        "mem_reservation": (ComposeByteValue, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#mem_reservation"),
        "secrets": (Secrets, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#secrets"),
        "scale": (int, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#scale"),
        "dns": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#dns"),
        "dns_opt": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#dns_opt"),
        "dns_search": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#dns_search"),
        "domainname": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#domainname"),
        "external_links": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#external_links"),
        "extra_hosts": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#extra_hosts"),
        "group_add": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#group_add"),
        "healthcheck": (HealthCheck.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#healthcheck"),
        "hostname": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#hostname"),
        "init": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#init"),
        "ipc": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#ipc"),
        "isolation": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#isolation"),
        "labels": (ComposeListOrMapElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#labels"),
        "links": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#links"),
        "logging": (Logging, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#logging"),
        "network_mode": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#network_mode"),
        "networks": (Networks, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#networks"),
        "mac_address": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#mac_address"),
        "mem_limit": (ComposeByteValue, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#mem_limit"),
        "mem_reservation": (ComposeByteValue, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#mem_reservation"),
        "mem_swappiness": ((int, [0, 100]), "https://github.com/compose-spec/compose-spec/blob/master/spec.md#mem_swappiness"),
        "memswap_limit": (ComposeByteValue, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#memswap_limit"),
        "oom_kill_disable": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#oom_kill_disable"),
        "oom_score_adj": ((int, [-1000, 1000]), "https://github.com/compose-spec/compose-spec/blob/master/spec.md#oom_score_adj"),
        "pid": (int, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#pid"),
        "pids_limit": (int, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#pids_limit"),
        "platform": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#platform"),
        "privileged": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#privileged"),
        "pull_policy": ((str, ["always", "never", "missing", "build"]), "https://github.com/compose-spec/compose-spec/blob/master/spec.md#pull_policy"),
        "read_only": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#read_only"),
        "restart": ((str, ["no", "always", "on-failure", "unless-stopped"]), "https://github.com/compose-spec/compose-spec/blob/master/spec.md#restart"),
        "runtime": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#runtime"),
        "security_opt": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#security_opt"),
        "shm_size": (ComposeByteValue, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#shm_size"),
        "stdin_open": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#stdin_open"),
        "stop_grace_period": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#stop_grace_period"),
        "stop_signal": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#stop_signal"),
        "storage_opt": (StorageOpt.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#storage_opt"),
        "sysctls": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#sysctls"),
        "tmpfs": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#tmpfs"),
        "tty": (bool, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#tty"),
        "ulimits": (Ulimits.from_parsed_yaml, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#ulimits"),
        "user": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#user"),
        "userns_mode": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#userns_mode"),
        "volumes": (Volumes, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#volumes"),
        "volumes_from": (ComposeStringOrListElement, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#volumes_from"),
        "working_dir": (str, "https://github.com/compose-spec/compose-spec/blob/master/spec.md#working_dir")

    }

    def entrypoint_and_command(self):
        if self.command is None and self.entrypoint is None:
            return None
        else:
            container_entrypoint_and_command = ""
            if self.entrypoint is not None:
                container_entrypoint_and_command += self.entrypoint.command_string()
                container_entrypoint_and_command += " "
            if self.command is not None:
                container_entrypoint_and_command += self.command.command_string()
            return container_entrypoint_and_command

    def resolve_environment_hierarchy(self):
        if self.env_file is not None:
            env_file = self.env_file.readFile()
            env_file.update(self.environment)
            return env_file
        else:
            return self.environment



from pycomposefile.compose_element import ComposeListOrMapElement, ComposeStringOrListElement


class Environment(ComposeListOrMapElement):
    pass


class EnvFile(ComposeStringOrListElement):

    def readFile(self):
        env_array = []

        for file_name in self:
            f = open(file_name, "r")

            for line in f.readlines():
                if not line.startswith("#"):
                    env_array.append(line.rstrip("\n"))
            f.close()
        return Environment(env_array)
